"use strict";

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it.return != null) it.return(); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

var _require = require('../util/math'),
    decimalAdjust = _require.decimalAdjust,
    isValidDecimal = _require.isValidDecimal,
    overflowsOrUnderflows = _require.overflowsOrUnderflows;

var ucum = require('ucum');

var Quantity = /*#__PURE__*/function () {
  function Quantity(value, unit) {
    _classCallCheck(this, Quantity);

    this.value = value;
    this.unit = unit;

    if (this.value == null || isNaN(this.value)) {
      throw new Error('Cannot create a quantity with an undefined value');
    } else if (!isValidDecimal(this.value)) {
      throw new Error('Cannot create a quantity with an invalid decimal value');
    } // Attempt to parse the unit with UCUM. If it fails, throw a friendly error.


    if (this.unit != null && !is_valid_ucum_unit(this.unit)) {
      throw new Error("'".concat(this.unit, "' is not a valid UCUM unit."));
    }
  }

  _createClass(Quantity, [{
    key: "clone",
    value: function clone() {
      return new Quantity(this.value, this.unit);
    }
  }, {
    key: "toString",
    value: function toString() {
      return "".concat(this.value, " '").concat(this.unit, "'");
    }
  }, {
    key: "sameOrBefore",
    value: function sameOrBefore(other) {
      if (other != null && other.isQuantity) {
        var other_v = convert_value(other.value, ucum_unit(other.unit), ucum_unit(this.unit));

        if (other_v == null) {
          return null;
        } else {
          return this.value <= other_v;
        }
      }
    }
  }, {
    key: "sameOrAfter",
    value: function sameOrAfter(other) {
      if (other != null && other.isQuantity) {
        var other_v = convert_value(other.value, ucum_unit(other.unit), ucum_unit(this.unit));

        if (other_v == null) {
          return null;
        } else {
          return this.value >= other_v;
        }
      }
    }
  }, {
    key: "after",
    value: function after(other) {
      if (other != null && other.isQuantity) {
        var other_v = convert_value(other.value, ucum_unit(other.unit), ucum_unit(this.unit));

        if (other_v == null) {
          return null;
        } else {
          return this.value > other_v;
        }
      }
    }
  }, {
    key: "before",
    value: function before(other) {
      if (other != null && other.isQuantity) {
        var other_v = convert_value(other.value, ucum_unit(other.unit), ucum_unit(this.unit));

        if (other_v == null) {
          return null;
        } else {
          return this.value < other_v;
        }
      }
    }
  }, {
    key: "equals",
    value: function equals(other) {
      if (other != null && other.isQuantity) {
        if (!this.unit && other.unit || this.unit && !other.unit) {
          return false;
        } else if (!this.unit && !other.unit) {
          return this.value === other.value;
        } else {
          var other_v = convert_value(other.value, ucum_unit(other.unit), ucum_unit(this.unit));

          if (other_v == null) {
            return null;
          } else {
            return decimalAdjust('round', this.value, -8) === decimalAdjust('round', other_v, -8);
          }
        }
      }
    }
  }, {
    key: "convertUnit",
    value: function convertUnit(to_unit) {
      var value = convert_value(this.value, this.unit, to_unit);
      var unit = to_unit; // Need to pass through constructor again to catch invalid units

      return new Quantity(value, unit);
    }
  }, {
    key: "dividedBy",
    value: function dividedBy(other) {
      return this.multiplyDivide(other, '/');
    }
  }, {
    key: "multiplyBy",
    value: function multiplyBy(other) {
      return this.multiplyDivide(other, '.'); // in ucum . represents multiplication
    }
  }, {
    key: "multiplyDivide",
    value: function multiplyDivide(other, operator) {
      if (other != null && other.isQuantity) {
        var a = this.unit != null ? this : new Quantity(this.value, '1');
        var b = other.unit != null ? other : new Quantity(other.value, '1');
        var can_val = a.to_ucum();
        var other_can_value = b.to_ucum();
        var ucum_value = ucum_multiply(can_val, [[operator, other_can_value]]);

        if (overflowsOrUnderflows(ucum_value.value)) {
          return null;
        }

        try {
          return new Quantity(ucum_value.value, units_to_string(ucum_value.units));
        } catch (e) {
          return null;
        }
      } else {
        var value = operator === '/' ? this.value / other : this.value * other;

        if (overflowsOrUnderflows(value)) {
          return null;
        }

        try {
          return new Quantity(decimalAdjust('round', value, -8), coalesceToOne(this.unit));
        } catch (e) {
          return null;
        }
      }
    }
  }, {
    key: "to_ucum",
    value: function to_ucum() {
      var u = ucum.parse(ucum_unit(this.unit));
      u.value *= this.value;
      return u;
    }
  }, {
    key: "isQuantity",
    get: function get() {
      return true;
    }
  }]);

  return Quantity;
}();

function clean_unit(units) {
  if (ucum_time_units[units]) {
    return ucum_to_cql_units[ucum_time_units[units]];
  } else {
    return units;
  }
} // Hash of time units and their UCUM equivalents, both case-sensitive and case-insensitive
// See http://unitsofmeasure.org/ucum.html#para-31
// The CQL specification says that dates are based on the Gregorian calendar
// UCUM says that years should be Julian. As a result, CQL-based year and month identifiers will
// be matched to the UCUM gregorian units. UCUM-based year and month identifiers will be matched
// to the UCUM julian units.


var ucum_time_units = {
  years: 'a_g',
  year: 'a_g',
  YEARS: 'a_g',
  YEAR: 'a_g',
  a_g: 'a_g',
  a: 'a_j',
  ANN: 'a_j',
  ann: 'a_j',
  A: 'a_j',
  a_j: 'a_j',
  months: 'mo_g',
  month: 'mo_g',
  mo_g: 'mo_g',
  mo: 'mo_j',
  MO: 'mo_j',
  mo_j: 'mo_j',
  weeks: 'wk',
  week: 'wk',
  wk: 'wk',
  WK: 'wk',
  days: 'd',
  day: 'd',
  d: 'd',
  D: 'd',
  hours: 'h',
  hour: 'h',
  h: 'h',
  H: 'h',
  minutes: 'min',
  minute: 'min',
  min: 'min',
  MIN: 'min',
  seconds: 's',
  second: 's',
  s: 's',
  S: 's',
  milliseconds: 'ms',
  millisecond: 'ms',
  ms: 'ms',
  MS: 'ms'
};
var ucum_to_cql_units = {
  a_j: 'year',
  a_g: 'year',
  mo_j: 'month',
  mo_g: 'month',
  wk: 'week',
  d: 'day',
  h: 'hour',
  min: 'minute',
  s: 'second',
  ms: 'millisecond'
}; // this is used to perform any conversions of CQL date time fields to their ucum equivalents

function ucum_unit(unit) {
  return ucum_time_units[unit] || unit || '';
} // just a wrapper function to deal with possible exceptions being thrown


function convert_value(value, from, to) {
  try {
    if (from === to) {
      return value;
    } else {
      return decimalAdjust('round', ucum.convert(value, ucum_unit(from), ucum_unit(to)), -8);
    } // If the units could not be alignied ie: incompareable, exception will be thrown, return null

  } catch (e) {
    return null;
  }
} // Cache for unit validity results so we dont have to go to ucum.js for every check.
// Is a map of unit string to boolean validity


var unitValidityCache = {}; // Helper for checking if a unit is valid. Checks the cache first, checks with ucum.js otherwise.

function is_valid_ucum_unit(unit) {
  if (unitValidityCache[unit] != null) {
    return unitValidityCache[unit];
  } else {
    try {
      ucum.parse(ucum_unit(unit));
      unitValidityCache[unit] = true;
      return true;
    } catch (error) {
      unitValidityCache[unit] = false;
      return false;
    }
  }
} // This method will take a ucum.js representation of units and convert them to a string
// ucum.js units are a has of unit => power values.  For instance m/h (meters per hour) in
// ucum.js will be reprsented by the json object {m: 1, h:-1}  negative values are inverted and
// are akin to denominator values in a fraction.  Positive values are somewhat a kin to numerator
// values in that they preceed the inverted values.  It is possible in ucum to have multiple non inverted
// or inverted values.  This method combines all of the non inverted values and appends them with
// the ucum multiplication operator '.' and then appends the inverted values separated by the ucum
// divisor '/' .


function units_to_string() {
  var units = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  var numer = [];
  var denom = [];

  for (var _i = 0, _Object$keys = Object.keys(units); _i < _Object$keys.length; _i++) {
    var key = _Object$keys[_i];
    var v = units[key];
    var pow = Math.abs(v);
    var str = pow === 1 ? key : key + pow;

    if (v < 0) {
      denom.push(str);
    } else {
      numer.push(str);
    }
  }

  var unit_string = '';
  unit_string += numer.join('.');

  if (denom.length > 0) {
    unit_string += '/' + denom.join('/');
  }

  if (unit_string === '') {
    return '1';
  } else {
    return unit_string;
  }
} // this method is taken from the ucum.js library which it does not  export
// so we need to replicate the behavior here in order to perform multiplication
// and division of the ucum values.
// t:  the ucum quantity being multiplied/divided .  This method modifies the object t that is passed in
// ms: an array of arrays whoes format is [<operator>,<ucum quantity>] an example would be [['.', {value: 1, units: {m:2}}]]
// this would represent multiply t by the value m^2


function ucum_multiply(t) {
  var ms = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];

  if (ms.length === 0) {
    return t;
  }

  var ret = t;

  var _iterator = _createForOfIteratorHelper(ms),
      _step;

  try {
    for (_iterator.s(); !(_step = _iterator.n()).done;) {
      var mterm = _step.value;
      var sign = mterm[0] === '.' ? 1 : -1;
      var b = mterm[1];
      ret.value *= Math.pow(b.value, sign);

      for (var k in b.units) {
        var v = b.units[k];
        ret.units[k] = ret.units[k] || 0;
        ret.units[k] = ret.units[k] + sign * v;

        if (ret.units[k] === 0) {
          delete ret.units[k];
        }
      }
    }
  } catch (err) {
    _iterator.e(err);
  } finally {
    _iterator.f();
  }

  return ret;
}

function parseQuantity(str) {
  var components = /([+|-]?\d+\.?\d*)\s*('(.+)')?/.exec(str);

  if (components != null && components[1] != null) {
    var value = parseFloat(components[1]);

    if (!isValidDecimal(value)) {
      return null;
    }

    var unit;

    if (components[3] != null) {
      unit = components[3].trim();
    } else {
      unit = '';
    }

    return new Quantity(value, unit);
  } else {
    return null;
  }
}

function doScaledAddition(a, b, scaleForB) {
  var b_unit;

  if (a != null && a.isQuantity && b != null && b.isQuantity) {
    var a_unit;
    var _ref = [coalesceToOne(a.unit), coalesceToOne(b.unit)];
    a_unit = _ref[0];
    b_unit = _ref[1];
    // The units don't have to match (m and m^2), but must be convertable
    // we will choose the unit of a to be the unit we return
    var val = convert_value(b.value * scaleForB, b_unit, a_unit);

    if (val == null) {
      return null;
    }

    var sum = a.value + val;

    if (overflowsOrUnderflows(sum)) {
      return null;
    } else {
      return new Quantity(sum, a_unit);
    }
  } else if (a.copy && a.add) {
    b_unit = b != null && b.isQuantity ? coalesceToOne(b.unit) : b.unit;
    return a.copy().add(b.value * scaleForB, clean_unit(b_unit));
  } else {
    throw new Error('Unsupported argument types.');
  }
}

function doAddition(a, b) {
  return doScaledAddition(a, b, 1);
}

function doSubtraction(a, b) {
  return doScaledAddition(a, b, -1);
}

function doDivision(a, b) {
  if (a != null && a.isQuantity) {
    return a.dividedBy(b);
  }
}

function doMultiplication(a, b) {
  if (a != null && a.isQuantity) {
    return a.multiplyBy(b);
  } else {
    return b.multiplyBy(a);
  }
}

function coalesceToOne(o) {
  if (o == null || o.trim != null && !o.trim()) {
    return '1';
  } else {
    return o;
  }
}

function compare_units(unit_a, unit_b) {
  try {
    var c = ucum.convert(1, ucum_unit(unit_a), ucum_unit(unit_b));

    if (c > 1) {
      return 1; // unit_a is bigger (less precise)
    }

    if (c < 1) {
      return -1; // unit_a is smaller
    }

    return 0; //units are the same
  } catch (e) {
    return null;
  }
}

module.exports = {
  Quantity: Quantity,
  convert_value: convert_value,
  parseQuantity: parseQuantity,
  doAddition: doAddition,
  doSubtraction: doSubtraction,
  doDivision: doDivision,
  doMultiplication: doMultiplication,
  compare_units: compare_units
};